﻿using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core;
using OpenTK.Graphics.OpenGL4;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Common.Input;
using OpenTK.Windowing.GraphicsLibraryFramework;

namespace GraphicsManager.Objects;

public class TexturedTextBox : IRenderObject
{
	private Rectangle[] _bounds;
	private Label _label;
	private Label _watermark;
	public ContextMenu? ContextMenu { get; set; }

	public TextLocation TextLocation { get; set; } = TextLocation.TopLeft;
	
	public TexturedTextBox(Texture Left, Texture Middle, Texture Right, FontFamily LabelFam, FontFamily WaterFam)
	{
		_label = new Label(LabelFam);
		_watermark = new(WaterFam)
		{
			Color = new(128, 128, 128, 255) 
		};
		_bounds = new Rectangle[] {new(Left), new(Middle), new(Right) };
	}
	
	public MouseCursor HoverMouse
	{
		get => _bounds[0].HoverMouse;
		set
		{
			for (int i = 0; i < _bounds.Length; i++)
			{
				_bounds[i].HoverMouse = value;
			}

			_watermark.HoverMouse = value;
			_label.HoverMouse = value;
		}
	}
	
	public int Border { get; set; } = 2;
	
	public TexturedTextBox(Texture Left, Texture Middle, Texture Right, FontInteraction LabelFam, FontInteraction WaterFam)
	{
		_label = new Label(LabelFam);
		_watermark = new(WaterFam)
		{
			Color = new(128, 128, 128, 255) 
		};
		_bounds = new Rectangle[] {new(Left), new(Middle), new(Right) };
	}

	private bool mouseinside;
	
	private void WindowOnMouseMove(MouseMoveEventArgs e)
	{
		if (Visible &&
		    Parent?.IntToFloat(Location.X) <= Window?.IntToFloat((float)Window?.MousePosition.X!) &&
		    Parent?.IntToFloat(Size.X + Location.X) >= Window?.IntToFloat((float)Window?.MousePosition.X!) &&
		    Parent?.IntToFloat(Location.Y + Size.Y, true) <= Window?.IntToFloat((float)Window?.MousePosition.Y!, true) &&
		    Parent?.IntToFloat(Location.Y, true) >= Window?.IntToFloat((float)Window?.MousePosition.Y!, true))
		{
			if (!mouseinside)
			{
				if (Window!.CurrentTop is null || Window!.CurrentTop == this)
				{
					Window!.Cursor = HoverMouse;
					if (MouseEnter is not null) _ = MouseEnter.Invoke(this);
					mouseinside = true;
				}
				else
				{
					mouseinside = false;
				}
			}
		}
		else
		{
			if (mouseinside)
			{
				if (mouseinside && Window!.Cursor == HoverMouse) Window!.Cursor = Parent!.HoverMouse;
				if (MouseLeave is not null && mouseinside) _ = MouseLeave.Invoke(this);
				mouseinside = false;
			}
		}
	}

	//private int CurrentIndex = 0;
	
	public event Func<IRenderObject, Task>? WindowLoaded;
	public event Func<IRenderObject, Task>? MouseEnter;
	public event Func<IRenderObject, Task>? MouseLeave;
	public object? Tag { get; set; } = null;
	public ObjectAnchor Anchor { get; set; }
	public FontInteraction Font { get => _label.Font; }
	public string Text
	{
		get => _label.Text;
		set
		{
			int old = _label.TrueHeight;
			_label.Text = value;
			if (!string.IsNullOrEmpty(value))
			{
				bool f = false;
				if (!_label.Visible)
				{
					f = true;
					_label.Visible = true;
					_label.Location = TextLocation switch
					{
						TextLocation.TrueCenterLeft => new(Location.X + Border + 5, Location.Y + ((Size.Y - _label.TrueHeight) / 2) - (_label.Size.Y  - _label.TrueHeight)),
						TextLocation.PostiveTureCenterLeft => new(Location.X + Border + 5, Location.Y + ((Size.Y - _label.PostiveTrueHeight) / 2) - _label.Size.Y + _label.TrueHeight),
						TextLocation.PxLeft => new(Location.X + Border + 5, Location.Y + ((Size.Y - _label.Size.Y) / 2)),
						_ => new(Location.X + Border + 5, Location.Y + Border + 5)
					};
					_watermark.Location = _label.Location;
				}
				if (_watermark.Visible) _watermark.Visible = false;
				if (!f && TextLocation == TextLocation.TrueCenterLeft && old != _label.TrueHeight)
				{
					_label.Location = new(Location.X + Border + 5, Location.Y + ((Size.Y - _label.TrueHeight) / 2) - (_label.Size.Y - _label.TrueHeight));
					_watermark.Location = _label.Location;
				}
			}
			else
			{
				if (_label.Visible) _label.Visible = false;
				if (!_watermark.Visible)
				{
					_watermark.Visible = true;
					_watermark.Location = TextLocation switch
					{
						TextLocation.TrueCenterLeft => new(Location.X + Border + 5, Location.Y + ((Size.Y - _watermark.TrueHeight) / 2) - (_watermark.Size.Y  - _watermark.TrueHeight)),
						TextLocation.PostiveTureCenterLeft => new(Location.X + Border + 5, Location.Y + ((Size.Y - _label.PostiveTrueHeight) / 2) - _label.Size.Y + _label.TrueHeight),
						TextLocation.PxLeft => new(Location.X + Border + 5, Location.Y + ((Size.Y - _watermark.Size.Y) / 2)),
						_ => new(Location.X + Border + 5, Location.Y + Border + 5)
					};
					_label.Location = _label.Location;
				}
			}
		}
	}
	
	public void ForceDistanceUpdate()
	{
		if (Parent is null) return;
		ForceDistanceUpdate(Parent);
	}
	
	public void ForceDistanceUpdate(IParent parent)
	{
		Distance = new(parent.Size.X - Size.X - Location.X, parent.Size.Y - Size.Y - Location.Y);
	}

	public FontInteraction WatermarkFont { get => _watermark.Font; }
	public string WatermarkText
	{
		get
		{
			return _watermark.Text;
		}
		set
		{
			_watermark.Text = value;
		}
	}
	public char? PasswordChar { get => _label.PasswordChar; set => _label.PasswordChar = value; }
	public bool Loaded { get; private set; }
	private Vector2i _size;
	public Vector2i Size
	{
		get
		{
			return _size;
		}
		set
		{
			if (_size == value) return;
			_bounds[0].Size = new(value.Y);
			_bounds[1].Size = new(value.X - (_bounds[0].Size.X * 2), value.Y);
			_bounds[2].Size = _bounds[0].Size;
			if (_size.X != value.X)
			{
				_bounds[1].Location = new(value.Y, 0);
				_bounds[2].Location = new(_bounds[1].Size.X + _bounds[1].Location.X, 0);
			}
			_size = value;
		}
	}
	public Vector2i Location { 
		get => _bounds[0].Location; 
		set
		{
			Vector2i diff = _bounds[0].Location - value;
			for (int i = 0; i < _bounds.Length; i++)
			{
				_bounds[i].Location -= diff;
			}
			if (_watermark.Visible && !string.IsNullOrEmpty(_watermark.Text))
			{
				_watermark.Location = TextLocation switch
				{
					TextLocation.TrueCenterLeft => new(value.X + Border + 5, value.Y + ((Size.Y - _watermark.TrueHeight) / 2) - (_watermark.Size.Y - _watermark.TrueHeight)),
					TextLocation.PostiveTureCenterLeft => new(value.X + Border + 5, value.Y + ((Size.Y - _label.PostiveTrueHeight) / 2) - _label.Size.Y + _label.TrueHeight),
					TextLocation.PxLeft => new(value.X + Border + 5, value.Y + ((Size.Y - _watermark.Size.Y) / 2)),
					_ => new(value.X + Border + 5, value.Y + Border + 5)
				};
				_label.Location = _watermark.Location;
			}
			else
			{
				_label.Location = TextLocation switch
				{
					TextLocation.TrueCenterLeft => new(value.X + Border + 5, value.Y + ((Size.Y - _label.TrueHeight) / 2) - (_label.Size.Y - _label.TrueHeight)),
					TextLocation.PostiveTureCenterLeft => new(value.X + Border + 5, value.Y + ((Size.Y - _label.PostiveTrueHeight) / 2) - _label.Size.Y + _label.TrueHeight),
					TextLocation.PxLeft => new(value.X + Border + 5, value.Y + ((Size.Y - _label.Size.Y) / 2)),
					_ => new(value.X + Border + 5, value.Y + Border + 5)
				};
				_watermark.Location = _label.Location;
			}
		}
	}

	public Vector2 SizeAsFloat { get; } = new();
	public Vector2 LocationAsFloat { get => _bounds[0].LocationAsFloat; }
	public Vector2i Distance { get; set; }
	public IParent? Parent { get; private set; }
	public Window? Window { get; private set; }
	public Color4 TextColor { get => _label.Color; set => _label.Color = value; }
	public Color4 WatermarkColor { get => _watermark.Color; set => _watermark.Color = value; }
	
	public bool Visible
	{
		get => _bounds[0].Visible;
		set 
		{
			if (value == _bounds[0].Visible) return;
			for (int i = 0; i < _bounds.Length; i++)
			{
				_bounds[i].Visible = value;
			}
			if (value)
			{
				if (!string.IsNullOrEmpty(_label.Text))
				{
					_label.Visible = true;
					_watermark.Visible = false;
				}
				else
				{
					_label.Visible = false;
					_watermark.Visible = true;
				}
			}
			else
			{
				_label.Visible = value;
				_watermark.Visible = value;
			}
		}
	}
	public event Func<IRenderObject, Task>? Clicked;

	public void Clean()
	{
		for (int i = 0; i < _bounds.Length; i++)
		{
			_bounds[i].Clean();
		}
		_label.Clean();
		_watermark.Clean();
	}

	//private Vector2i? l, s;

	public void Draw(int x, int y, int w, int h)
	{
		if (!Visible || !Loaded) return;
		for (int i = 0; i < _bounds.Length; i++)
		{
			_bounds[i].Draw(x,y,w,h);
		}
		int nx = x, ny = y, nw = w, nh = h;
		if (Location.X + Border > nw)
			return;
		else
		{
			nx += (Location.X + Border);
			nw -= (Location.X + Border);
			if (Size.X - Border < nw)
				nw = Size.X - Border;
		}

		if (Location.Y + Border > nh)
			return;
		else
		{
			ny += (Location.Y + Border);
			nh -= (Location.Y + Border);
			if (Size.Y - Border < nh)
				nh = Size.Y - Border;
		}
		if (nh < 1 || nw < 1) return;
		GL.Scissor(nx,ny,nw,nh);
		if (!string.IsNullOrEmpty(_label.Text)) _label.Draw(nx,ny,nw,nh);
		else _watermark.Draw(nx,ny,nw,nh);
	}

	public void LoadToParent(IParent parent, Window window)
	{
		if (Loaded) return;
		Parent = parent;
		Window = window;
		Window.MouseDown += Window_MouseDown;
		Window.KeyDown += Window_KeyDown;
		Window.TextInput += WindowOnTextInput;
		Loaded = true;
		if (!Window!.Context.IsCurrent) Window.Context.MakeCurrent();
		for (int i = 0; i < _bounds.Length; i++)
		{
			_bounds[i].LoadToParent(Parent, Window);
		}
		_label.LoadToParent(Parent, Window);
		_watermark.LoadToParent(Parent, Window);
		Window.MouseMove += WindowOnMouseMove;
		Location = Location;
		if (WindowLoaded is not null) WindowLoaded.Invoke(this);
	}

	private void WindowOnTextInput(TextInputEventArgs obj)
	{
		if (!use) return;
		Text += obj.AsString;
	}

	private bool use;
	public event Func<KeyboardKeyEventArgs, Task>? KeyPress;

	public void UnFocus()
	{
		use = false;
		if (Window is not null && Window.focused == this)
			Window.focused = null;
	}
	public void Focus()
	{
		if (Window is not null)
		{
			if (Window.focused is not null)
			{
				Window.focused.UnFocus();
			}

			Window.focused = this;
			use = true;
		}
	}
	private void Window_KeyDown(KeyboardKeyEventArgs obj)
	{
		if (!use) return;
		if (obj.Key == Keys.CapsLock || obj.Key == Keys.Menu || obj.Key == Keys.LeftSuper || obj.Key == Keys.RightSuper || obj.Key == Keys.End || obj.Key == Keys.Home || obj.Key == Keys.PageDown || obj.Key == Keys.PageUp || obj.Key == Keys.Insert || obj.Key == Keys.Up || obj.Key == Keys.Down || obj.Key == Keys.Left || obj.Key == Keys.Right) return;
		if (obj.Key == Keys.Backspace || obj.Key == Keys.Delete)
		{
			if (!(Text.Length > 0)) return;
			Text = Text.Remove(Text.Length - 1, 1);
		}

		if (obj.Key == Keys.V && obj.Control && Window is not null) Text += Window.ClipboardString;
		if (KeyPress is not null) _ = KeyPress.Invoke(obj);
	}

	private void Window_MouseDown(MouseButtonEventArgs e)
	{
		if (Visible &&
			e.Button == MouseButton.Button1 &&
			Parent?.IntToFloat(Location.X) <= Window?.IntToFloat((int)Window?.MousePosition.X!) &&
			Parent?.IntToFloat(Size.X + Location.X) >= Window?.IntToFloat((int)Window?.MousePosition.X!) &&
			Parent?.IntToFloat(Location.Y + Size.Y, true) <= Window?.IntToFloat((int)Window?.MousePosition.Y!, true) &&
			Parent?.IntToFloat(Location.Y, true) >= Window?.IntToFloat((int)Window?.MousePosition.Y!, true))
		{
			use = true;
			Focus();
			if (Clicked is not null) Clicked.Invoke(this);
		}
		else use = false;
	}
}
